{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "[Day 9](https://adventofcode.com/2019/day/9) puzzle input is IntCode program called \"BOOST\" which is a test that the IntCode CPU is operating correctly (mine is [here](puzzle_input/day9.txt)). Part 1 involves running the program with a given input until it succeeds and outputs a code. Part 2 involves running this program with a different input - it'll output another code."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[2351176124]\n"
     ]
    }
   ],
   "source": [
    "opcode_add = 1\n",
    "opcode_mul = 2\n",
    "opcode_read = 3\n",
    "opcode_write = 4\n",
    "opcode_jump_true = 5\n",
    "opcode_jump_false = 6\n",
    "opcode_lt = 7\n",
    "opcode_eq = 8\n",
    "opcode_rebase = 9\n",
    "\n",
    "opcode_terminate = 99\n",
    "\n",
    "mode_position = 0\n",
    "mode_immediate = 1\n",
    "mode_relative = 2\n",
    "\n",
    "class CpuState:\n",
    "    def __init__(self, memory, pc, offset):\n",
    "        self.memory = memory\n",
    "        self.pc = pc\n",
    "        self.offset = offset\n",
    "\n",
    "            \n",
    "def resize_memory(state, target_addr):\n",
    "    state.memory += ([0] * (1 + target_addr - len(state.memory)))\n",
    "            \n",
    "            \n",
    "def uop_read(state, value, mode):\n",
    "    debug_log(\"\\tuop_read\")\n",
    "    if mode == mode_position:\n",
    "        if value >= len(state.memory):\n",
    "            resize_memory(state, value)        \n",
    "        \n",
    "        return int(state.memory[value])\n",
    "    \n",
    "    elif mode == mode_relative:\n",
    "        if state.offset + value >= len(state.memory):\n",
    "            resize_memory(state, state.offset + value)     \n",
    "        \n",
    "        return int(state.memory[state.offset + value])\n",
    "    \n",
    "    elif mode == mode_immediate:\n",
    "        return int(value)\n",
    "    \n",
    "    else:\n",
    "        raise Exception(\"UNKNOWN MODE\")\n",
    "\n",
    "def uop_write(state, dst, value, mode):\n",
    "    debug_log(\"\\tuop_write\")\n",
    "    if mode == mode_position:\n",
    "        if dst >= len(state.memory):\n",
    "            resize_memory(state, dst)\n",
    "        state.memory[dst] = value\n",
    "        \n",
    "    elif mode == mode_relative:\n",
    "        if state.offset + dst >= len(state.memory):\n",
    "            resize_memory(state, state.offset + dst)     \n",
    "        \n",
    "        state.memory[state.offset + dst] = value\n",
    "            \n",
    "    elif mode == mode_immediate:\n",
    "        raise Exception(f\"cannot write {value} to literal {dst}\")\n",
    "    \n",
    "    \n",
    "def uop_cond_jump(state, modes, cond):\n",
    "    param_mode = try_pop_mode(modes)\n",
    "    param_raw = int(state.memory[state.pc + 1])\n",
    "    param = uop_read(state, param_raw, param_mode)\n",
    "    \n",
    "    dest_mode = try_pop_mode(modes)\n",
    "    dest_raw = int(state.memory[state.pc + 2])\n",
    "    dest = uop_read(state, dest_raw, dest_mode)\n",
    "    \n",
    "    debug_log(\"\\tuop_cond_jump\", param, dest)\n",
    "    \n",
    "    if cond(param):\n",
    "        return dest\n",
    "    \n",
    "    return state.pc + 3\n",
    "    \n",
    "def uop_cmp(state, modes, cmp):\n",
    "    \n",
    "    param0_mode = try_pop_mode(modes)\n",
    "    param0_raw = int(state.memory[state.pc + 1])\n",
    "    param0 = uop_read(state, param0_raw, param0_mode)\n",
    "    \n",
    "    param1_mode = try_pop_mode(modes)\n",
    "    param1_raw = int(state.memory[state.pc + 2])\n",
    "    param1 = uop_read(state, param1_raw, param1_mode)\n",
    "    \n",
    "    dest_mode = try_pop_mode(modes)\n",
    "    dest = int(state.memory[state.pc + 3])\n",
    "    \n",
    "    debug_log(\"\\tuop_cmp\", param0, param1, dest, dest_mode)\n",
    "    \n",
    "    if cmp(param0, param1):\n",
    "        uop_write(state, dest, 1, dest_mode)\n",
    "    else:\n",
    "        uop_write(state, dest, 0, dest_mode)\n",
    "    \n",
    "    return state.pc + 4    \n",
    "    \n",
    "    \n",
    "def op_add(state, modes, input_buffer, output_buffer):\n",
    "    debug_log(\"op_add\")\n",
    "    arg0_mode = try_pop_mode(modes)\n",
    "    arg1_mode = try_pop_mode(modes)\n",
    "    dest_mode = try_pop_mode(modes)\n",
    "    \n",
    "    arg0_raw = int(state.memory[state.pc + 1])\n",
    "    arg1_raw = int(state.memory[state.pc + 2])\n",
    "    dest = int(state.memory[state.pc + 3])\n",
    "    \n",
    "    arg0 = uop_read(state, arg0_raw, arg0_mode)\n",
    "    arg1 = uop_read(state, arg1_raw, arg1_mode)\n",
    "    \n",
    "    debug_log(\"\\t\", dest, \"=\", arg0, \"+\", arg1)\n",
    "    uop_write(state, dest, str(int(arg0) + int(arg1)), dest_mode)\n",
    "    return state.pc + 4\n",
    "\n",
    "def op_mul(state, modes, input_buffer, output_buffer):\n",
    "    debug_log(\"op_mul\")\n",
    "    arg0_mode = try_pop_mode(modes)\n",
    "    arg1_mode = try_pop_mode(modes)\n",
    "    dest_mode = try_pop_mode(modes)\n",
    "    \n",
    "    arg0_raw = int(state.memory[state.pc + 1])\n",
    "    arg1_raw = int(state.memory[state.pc + 2])\n",
    "    dest = int(state.memory[state.pc + 3])\n",
    "    \n",
    "    arg0 = uop_read(state, arg0_raw, arg0_mode)\n",
    "    arg1 = uop_read(state, arg1_raw, arg1_mode)\n",
    "    \n",
    "    debug_log(\"\\t\", dest, \"=\", arg0, \"*\", arg1)\n",
    "    uop_write(state, dest, str(int(arg0) * int(arg1)), dest_mode)    \n",
    "    return state.pc + 4\n",
    "\n",
    "\n",
    "\n",
    "# Opcode 3 takes a single integer as input and \n",
    "# saves it to the position given by its only parameter. For example, the instruction 3,50 \n",
    "# would take an input value and store it at address 50.\n",
    "def op_read(state, modes, input_buffer, output_buffer):\n",
    "    debug_log(\"op_read\")  \n",
    "    \n",
    "    dest_mode = try_pop_mode(modes)\n",
    "    dest = int(state.memory[state.pc + 1])\n",
    "    val = input_buffer.pop()\n",
    "    debug_log(\"\\tdest\", dest)\n",
    "    debug_log(\"\\tval\", val)\n",
    "    \n",
    "    uop_write(state, dest, str(val), dest_mode)\n",
    "    debug_log(\"\\tread:\", state.memory[dest])\n",
    "    \n",
    "    return state.pc + 2\n",
    "\n",
    "# Opcode 4 outputs the value of its only parameter. For example, the instruction 4,50 \n",
    "# would output the value at address 50.\n",
    "def op_write(state, modes, input_buffer, output_buffer):\n",
    "    debug_log(\"op_write\")\n",
    "    \n",
    "    src_mode = try_pop_mode(modes)\n",
    "    src_raw = int(state.memory[state.pc + 1])\n",
    "    src = uop_read(state, src_raw, src_mode)\n",
    "    \n",
    "    output_buffer.append(src)\n",
    "    \n",
    "    return state.pc + 2\n",
    "\n",
    "\n",
    "# Opcode 5 is jump-if-true: \n",
    "#    if the first parameter is non-zero, it sets the instruction pointer to \n",
    "#    the value from the second parameter. Otherwise, it does nothing.\n",
    "def op_jump_true(state, modes, input_buffer, output_buffer):\n",
    "    debug_log(\"op_jump_true\")\n",
    "    return uop_cond_jump(state, modes, lambda x: x != 0)\n",
    "\n",
    "\n",
    "# Opcode 6 is jump-if-false: \n",
    "#    if the first parameter is zero, it sets the instruction pointer to the\n",
    "#    value from the second parameter. Otherwise, it does nothing.\n",
    "def op_jump_false(state, modes, input_buffer, output_buffer):\n",
    "    debug_log(\"op_jump_false\")\n",
    "\n",
    "    return uop_cond_jump(state, modes, lambda x: x == 0)\n",
    "\n",
    "# Opcode 7 is less than: \n",
    "#    if the first parameter is less than the second parameter, it stores 1 \n",
    "#    in the position given by the third parameter. Otherwise, it stores 0.\n",
    "def op_lt(state, modes, input_buffer, output_buffer):\n",
    "    debug_log(\"op_lt\")\n",
    "    \n",
    "    return uop_cmp(state, modes, lambda x, y: x < y)\n",
    "\n",
    "# Opcode 8 is equals: \n",
    "#    if the first parameter is equal to the second parameter, it stores 1 \n",
    "#    in the position given by the third parameter. Otherwise, it stores 0.\n",
    "def op_eq(state, modes, input_buffer, output_buffer):\n",
    "    debug_log(\"op_eq\")\n",
    "    \n",
    "    return uop_cmp(state, modes, lambda x, y: x == y)\n",
    "\n",
    "# Opcode 9 adjusts the relative base by the value of its only parameter. \n",
    "#     The relative base increases (or decreases, if the value is negative) \n",
    "#     by the value of the parameter.\n",
    "def op_rebase(state, modes, input_buffer, output_buffer):    \n",
    "    param_mode = try_pop_mode(modes)\n",
    "    param_raw = int(state.memory[state.pc + 1])\n",
    "    param = uop_read(state, param_raw, param_mode)\n",
    "\n",
    "    state.offset += param\n",
    "\n",
    "    return state.pc + 2\n",
    "\n",
    "\n",
    "def decode_op(instr):\n",
    "    if len(instr) > 2:\n",
    "        return int(instr[-2:])\n",
    "    return int(instr)\n",
    "    \n",
    "def decode_modes(instr):\n",
    "    if len(instr) > 2:\n",
    "        return [ int(d) for d in instr[:-2]]\n",
    "    return []\n",
    "    \n",
    "def try_pop_mode(modes):\n",
    "    if len(modes) == 0:\n",
    "        return 0\n",
    "    \n",
    "    return modes.pop()\n",
    "\n",
    "\n",
    "opcodes = {\n",
    "    opcode_add: op_add,\n",
    "    opcode_mul: op_mul,\n",
    "    opcode_read: op_read,\n",
    "    opcode_write: op_write,\n",
    "    opcode_jump_true: op_jump_true,\n",
    "    opcode_jump_false: op_jump_false,\n",
    "    opcode_lt: op_lt,\n",
    "    opcode_eq: op_eq,\n",
    "    opcode_rebase: op_rebase\n",
    "}\n",
    "\n",
    "def decode_instr(instr):\n",
    "    instr = str(instr)\n",
    "    opcode = decode_op(instr)\n",
    "    modes = decode_modes(instr)\n",
    "    \n",
    "    if not (opcode in opcodes):\n",
    "        raise Exception(f\"Invalid opcode {opcode}\")\n",
    "    \n",
    "    return (opcodes[opcode], modes)\n",
    "\n",
    "debug_on = False\n",
    "\n",
    "def debug_log(*args):\n",
    "    global debug_on     \n",
    "    if debug_on:\n",
    "        print(*args)\n",
    "\n",
    "        \n",
    "def run_intcode(program, input_buffer, output_buffer):\n",
    "    state = CpuState(program, 0, 0)\n",
    "    instr = state.memory[state.pc]\n",
    "    \n",
    "    while int(instr) != opcode_terminate:\n",
    "        debug_log(state.pc, state.offset, instr)\n",
    "        (op, modes) = decode_instr(instr)\n",
    "        state.pc = op(state, modes, input_buffer, output_buffer)\n",
    "        instr = state.memory[state.pc]\n",
    "        debug_log()\n",
    "        \n",
    "    return program[0]\n",
    "\n",
    "\n",
    "def run_program(program, input_buffer, output_buffer, noun=None, verb=None):\n",
    "    program_copy = [ x for x in program ]\n",
    "    \n",
    "    if noun:\n",
    "        program_copy[1] = noun\n",
    "    if verb:\n",
    "        program_copy[2] = verb\n",
    "        \n",
    "    return run_intcode(program_copy, input_buffer, output_buffer)\n",
    "\n",
    "def calculate_keycode(program, input_buffer):\n",
    "    output_buffer = []\n",
    "    run_program(program, input_buffer, output_buffer)\n",
    "    print(output_buffer)\n",
    "\n",
    "puzzle_input = open(\"puzzle_input/day9.txt\", \"r\").read().split(\",\")\n",
    "calculate_keycode(puzzle_input, [1])"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[73110]\n"
     ]
    }
   ],
   "source": [
    "calculate_keycode(puzzle_input, [2])"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.9.0"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
